/*****************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    classifyImages.cpp
** AUTHOR(S):   Paul Baumstarck <pbaumstarck@stanford.edu>
**
** DESCRIPTION:
**  Application for verifying Cuda functions.
**
*****************************************************************************/

#include <cstdlib>
#include <cassert>
#include <iostream>
#include <fstream>
#include <iomanip>
#include <vector>
#include <string.h>

#include "cv.h"
#include "cxcore.h"
#include "highgui.h"

#include "svlBase.h"
#include "svlML.h"
#include "svlVision.h"
#include "svlCuda.h"

using namespace std;

#define WINDOW_NAME "cudaVerify"
#define MULTI_PASS
#define PASSES 1
#define TRACK_ALLOC 0
#define PRINT_NAMED_STATUS(name,status) printf("%s -- ",name); if (status) { printf("PASSED\n"); } else { printf("*** FAILED ***\n"); }
#define PRINT_STATUS PRINT_NAMED_STATUS(test_names.back(),gb)
#define TRANFER_TOL 1e-8

#define PRINT_TOL 0

#define TEST_ALLOCATION
#define TEST_MATH
#define TEST_MINMAX
#define TEST_INTEGRAL
#define TEST_CONVOLUTION
#define TEST_FILTERING
#define TEST_DECISION_TREES
#define TEST_SLIDING_WINDOW

void minMax_gold(const float *data, float *minf, float *maxf, int w, int h, int mw = 4, int mh = 4);
void mean_gold(const float *data, float *outf, int w, int h, int mw = 32, int mh = 32);
void integralLine_gold(const float *data, float *integ, int w, int dim, bool do_cumsum);
void integralBlock_gold(const float *data, float *integ, float *integ2, int w, int h, int dim);
void rowIntegralBlock_gold(const float *data, float *integ, float *integ2, int w, int h, int dim);
void sub2ind(CvSize sz, const vector<CvPoint> &pts, vector<int> &ret);
void sortix(const vector<int> &vec, vector<int> &ix);
void conv2d_same(const float *data, int dw, int dh, const float *pat, int pw, int ph, float *res);
void conv2d_full(const float *data, int dw, int dh, const float *pat, int pw, int ph, float *res);
void ccoeff_normed_gold(const float *data, int dw, int dh, const float *pat, int pw, int ph, float *res);
void sobel_gold(const float *data, int dw, int dh, float *res);

void scaleLocsFeats2File(const char *fname, const ScaleLocations &locs, const ScaleFeatures &feats);
void scaleLocsFeatsD2File(const char *fname, const ScaleLocations &locs, const ScaleFeaturesD &feats);

int main(int argc, char *argv[])
{
#if 0
	// Test nested timer.
	svlLoopTimer t;
	int i_trials = 10;
	t.push(i_trials);
	for ( int i=0; i<i_trials; ++i ) {
		/*int j_trials = 5;
		t.push(j_trials);
		for ( int j=0; j<j_trials; ++j ) {*/
			t.printETC();
			clock_t tic = clock();
			while ( clock()-tic < CLOCKS_PER_SEC ) {}
		/*	t.inc();
		}
		t.pop();*/
		t.inc();
	}
	t.printETC();
	t.clear();
	printf("Again:\n");
	t.printETC();
#elif 0
	svlCudaMemHandler *cm;
	svlCudaPitch *untracked = cm->allocPitch(640, 480);

	svlCudaMemHandler::setTracking(true); // Enables tracking of allocations.
	svlCudaPitch *tracked_unnamed = cm->allocPitch(640, 480);
	svlCudaPitch *tracked_named = cm->allocPitch(640, 480, "named pitch");

	svlCudaMemHandler::printAllAllocations();

#else
	svlCudaMemHandler *cm;
	cm->setTracking(true);
	int ws[] = {201, 640};
	int hs[] = {201, 480};
	int zs[] = {5, 3};
	int szs[] = {ws[0]*hs[0], ws[1]*hs[1]};
	int sz3s[] = {ws[0]*hs[0]*zs[0], ws[1]*hs[1]*zs[1]};
	srand(time(NULL));

	// Random data.
	float *datas[] = { (float*)malloc(sizeof(float)*szs[0]),
					   (float*)malloc(sizeof(float)*szs[1]) };
	float *datas_2[] = { (float*)malloc(sizeof(float)*szs[0]),
					     (float*)malloc(sizeof(float)*szs[1]) };
	float *tests[] = { (float*)malloc(sizeof(float)*szs[0]),
					   (float*)malloc(sizeof(float)*szs[1]) };
	float *verfs[] = { (float*)malloc(sizeof(float)*szs[0]),
					   (float*)malloc(sizeof(float)*szs[1]) };
	float *data3s[] = { (float*)malloc(sizeof(float)*sz3s[0]),
					    (float*)malloc(sizeof(float)*sz3s[1]) };
	float *data3s_2[] = { (float*)malloc(sizeof(float)*sz3s[0]),
					      (float*)malloc(sizeof(float)*sz3s[1]) };
	float *test3s[] = { (float*)malloc(sizeof(float)*sz3s[0]),
					    (float*)malloc(sizeof(float)*sz3s[1]) };
	float *verf3s[] = { (float*)malloc(sizeof(float)*sz3s[0]),
					    (float*)malloc(sizeof(float)*sz3s[1]) };
	for ( unsigned j=0; j<2; ++j ) {
		for ( unsigned i=0; i<szs[j]; ++i ) {
			datas[j][i] = float(rand()) / float(RAND_MAX);
			datas_2[j][i] = float(rand()) / float(RAND_MAX);
		}
		for ( unsigned i=0; i<sz3s[j]; ++i )
			data3s[j][i] = float(rand()) / float(RAND_MAX);
	}
	vector<const char*> test_names;
	vector<bool> passeds;
	bool gb;
	float gf;


#ifdef TEST_ALLOCATION
	// *** Test memory allocation and transfer.
	// Test lines.
	{
		test_names.push_back("Cuda line allocation and transfer");
		gb = true;
		for ( unsigned j=0; gb && j<2; ++j ) {
			svlCudaLine *ln = cm->allocLine(ws[j]);
			cm->lineToGPU(ln, datas[j]);
			cm->lineFromGPU(tests[j], ln);
			float max_d = 0.0, sum_d = 0.0;
			for ( unsigned i=0; i<ws[j]; ++i ) {
				sum_d += fabs( datas[j][i] - tests[j][i] );
				max_d = max( max_d, fabs( datas[j][i] - tests[j][i] ) );
			}
			if ( sum_d >= TRANFER_TOL || max_d >= TRANFER_TOL )
				gb = false;
			cm->freeLine(ln);
		}
		passeds.push_back(gb);
		PRINT_STATUS
	}

	// Test pitches.
	{
		test_names.push_back("Cuda pitch allocation and transfer");
		gb = true;
		for ( unsigned j=0; gb && j<2; ++j ) {
			svlCudaPitch *p = cm->allocPitch(ws[j], hs[j]);
			cm->pitchToGPU(p, datas[j]);
			cm->pitchFromGPU(tests[j], p);
			float max_d = 0.0, sum_d = 0.0;
			for ( unsigned i=0; i<szs[j]; ++i ) {
				sum_d += fabs( datas[j][i] - tests[j][i] );
				max_d = max( max_d, fabs( datas[j][i] - tests[j][i] ) );
			}
			if ( sum_d >= TRANFER_TOL || max_d >= TRANFER_TOL )
				gb = false;
			cm->freePitch(p);
		}
		passeds.push_back(gb);
		PRINT_STATUS
	}

	// Test 3d pitches.
	{
		test_names.push_back("Cuda 3d pitch allocation and transfer");
		gb = true;
		for ( unsigned j=0; gb && j<2; ++j ) {
			svlCudaPitch3d *p = cm->allocPitch3d(ws[j], hs[j], zs[j]);
			cm->pitch3dToGPU(p, data3s[j]);
			cm->pitch3dFromGPU(test3s[j], p);
			float max_d = 0.0, sum_d = 0.0;
			for ( unsigned i=0; i<sz3s[j]; ++i ) {
				sum_d += fabs( data3s[j][i] - test3s[j][i] );
				max_d = max( max_d, fabs( data3s[j][i] - test3s[j][i] ) );
			}
			if ( sum_d >= TRANFER_TOL || max_d >= TRANFER_TOL )
				gb = false;
			cm->freePitch3d(p);
		}
		passeds.push_back(gb);
		PRINT_STATUS
	}
#endif


#ifdef TEST_MATH
	// *** Test general math functions.
	// Test addition.
	{
		test_names.push_back("Pitch add");
		gb = true;
		int n_reps = 5;
		float gf;
		for ( unsigned j=0; gb && j<2; ++j ) {
			svlCudaPitch *A = cm->allocPitch(ws[j], hs[j]),
						 *B = cm->allocPitch(ws[j], hs[j]),
						 *C = cm->allocPitch(ws[j], hs[j]);
			
			// Test two-pitch destructive add.
			cm->pitchToGPU(B, datas_2[j]);
			for ( int k=0; gb && k<n_reps; ++k ) {
				cm->pitchToGPU(A, datas[j]);
				float fac = 2.0f * float(rand())/float(RAND_MAX) - 1.0f;
				// Performs <A += fac*B>.
				svlCudaMath::add(A, B, fac);

				// Compute  gold result.
				cm->pitchFromGPU(tests[j], A);
				float max_d = 0.0, mean_d = 0.0;
				for ( unsigned i=0; i<szs[j]; ++i ) {
					gf = datas[j][i] + fac*datas_2[j][i];
					mean_d += fabs( gf - tests[j][i] );
					max_d = max( max_d, fabs( gf - tests[j][i] ) );
				}
				mean_d /= float(A->size());
				//printf("mean: %0.9f; max: %0.9f\n", mean_d, max_d);
				if ( mean_d >= 1e-7 || max_d >= 1e-6 )
					gb = false;
			}

			// Test three-pitch add.
			cm->pitchToGPU(A, datas[j]);
			cm->pitchToGPU(B, datas_2[j]);
			for ( int k=0; gb && k<n_reps; ++k ) {
				float fac_a = 2.0f * float(rand())/float(RAND_MAX) - 1.0f,
					  fac_b = 2.0f * float(rand())/float(RAND_MAX) - 1.0f;
				// Performs <C = fac_a*A + fac_b*B>.
				svlCudaMath::add(C, A, B, fac_a, fac_b);

				// Compute  gold result.
				cm->pitchFromGPU(tests[j], C);
				float max_d = 0.0, mean_d = 0.0;
				for ( unsigned i=0; i<szs[j]; ++i ) {
					gf = fac_a * datas[j][i] + fac_b * datas_2[j][i];
					mean_d += fabs( gf - tests[j][i] );
					max_d = max( max_d, fabs( gf - tests[j][i] ) );
				}
				mean_d /= float(A->size());
				//printf("mean: %0.9f; max: %0.9f\n", mean_d, max_d);
				if ( mean_d >= 1e-7 || max_d >= 1e-6 )
					gb = false;
			}

			cm->freePitch(A);
			cm->freePitch(B);
			cm->freePitch(C);
		}
		passeds.push_back(gb);
		PRINT_STATUS
	}

	// Test multiplication.
	{
		test_names.push_back("Pitch multiply");
		gb = true;
		float gf;
		int n_reps = 5;
		for ( unsigned j=0; gb && j<2; ++j ) {
			svlCudaPitch *A = cm->allocPitch(ws[j], hs[j]),
						 *B = cm->allocPitch(ws[j], hs[j]),
						 *C = cm->allocPitch(ws[j], hs[j]);
			
			// Test two-pitch destructive mult.
			cm->pitchToGPU(B, datas_2[j]);
			for ( int k=0; gb && k<n_reps; ++k ) {
				cm->pitchToGPU(A, datas[j]);
				float fac = 2.0f * float(rand())/float(RAND_MAX) - 1.0f;
				// Performs <A *= fac*B>.
				svlCudaMath::mult(A, B, fac);

				// Compute  gold result.
				cm->pitchFromGPU(tests[j], A);
				float max_d = 0.0, mean_d = 0.0;
				for ( unsigned i=0; i<szs[j]; ++i ) {
					gf = datas[j][i] * fac*datas_2[j][i];
					mean_d += fabs( gf - tests[j][i] );
					max_d = max( max_d, fabs( gf - tests[j][i] ) );
				}
				mean_d /= float(A->size());
				//printf("mean: %0.9f; max: %0.9f\n", mean_d, max_d);
				if ( mean_d >= 1e-7 || max_d >= 1e-6 )
					gb = false;
			}

			// Test three-pitch mult.
			cm->pitchToGPU(A, datas[j]);
			cm->pitchToGPU(B, datas_2[j]);
			for ( int k=0; gb && k<n_reps; ++k ) {
				float fac = 2.0f * float(rand())/float(RAND_MAX) - 1.0f;
				// Performs <C = fac*A*B>.
				svlCudaMath::mult(C, A, B, fac);

				// Compute  gold result.
				cm->pitchFromGPU(tests[j], C);
				float max_d = 0.0, mean_d = 0.0;
				for ( unsigned i=0; i<szs[j]; ++i ) {
					gf = datas[j][i] * fac * datas_2[j][i];
					mean_d += fabs( gf - tests[j][i] );
					max_d = max( max_d, fabs( gf - tests[j][i] ) );
				}
				mean_d /= float(A->size());
				//printf("mean: %0.9f; max: %0.9f\n", mean_d, max_d);
				if ( mean_d >= 1e-7 || max_d >= 1e-6 )
					gb = false;
			}

			cm->freePitch(A);
			cm->freePitch(B);
			cm->freePitch(C);
		}
		passeds.push_back(gb);
		PRINT_STATUS
	}

	// Test division.
	{
		test_names.push_back("Pitch divide");
		gb = true;
		float gf;
		int n_reps = 5;
		for ( unsigned j=0; gb && j<2; ++j ) {
			svlCudaPitch *A = cm->allocPitch(ws[j], hs[j]),
						 *B = cm->allocPitch(ws[j], hs[j]),
						 *C = cm->allocPitch(ws[j], hs[j]);
			
			float *temp = (float*)malloc(sizeof(float)*szs[j]);
			memcpy(temp, datas_2[j], sizeof(float)*szs[j]);
			// Add 1 to datas_2 for stability with division.
			for ( unsigned i=0; i<szs[j]; ++i )
				temp[i] += 1.0f;
			cm->pitchToGPU(B, temp);

			// Test two-pitch destructive divide.
			for ( int k=0; gb && k<n_reps; ++k ) {
				cm->pitchToGPU(A, datas[j]);
				float fac = 2.0f * float(rand())/float(RAND_MAX) - 1.0f;
				// Performs <A *= fac/B>.
				svlCudaMath::div(A, B, fac);

				// Compute  gold result.
				cm->pitchFromGPU(tests[j], A);
				float max_d = 0.0, mean_d = 0.0;
				for ( unsigned i=0; i<szs[j]; ++i ) {
					gf = datas[j][i] * fac / temp[i];
					mean_d += fabs( gf - tests[j][i] );
					max_d = max( max_d, fabs( gf - tests[j][i] ) );
				}
				mean_d /= float(A->size());
				//printf("mean: %0.9f; max: %0.9f\n", mean_d, max_d);
				if ( mean_d >= 1e-7 || max_d >= 1e-6 )
					gb = false;
			}

			// Test three-pitch divide.
			cm->pitchToGPU(A, datas[j]);
			for ( int k=0; gb && k<n_reps; ++k ) {
				float fac = 2.0f * float(rand())/float(RAND_MAX) - 1.0f;
				// Performs <C = fac*A/B>.
				svlCudaMath::div(C, A, B, fac);

				// Compute  gold result.
				cm->pitchFromGPU(tests[j], C);
				float max_d = 0.0, mean_d = 0.0;
				for ( unsigned i=0; i<szs[j]; ++i ) {
					gf = datas[j][i] * fac / temp[i];
					mean_d += fabs( gf - tests[j][i] );
					max_d = max( max_d, fabs( gf - tests[j][i] ) );
				}
				mean_d /= float(A->size());
				//printf("mean: %0.9f; max: %0.9f\n", mean_d, max_d);
				if ( mean_d >= 1e-7 || max_d >= 1e-6 )
					gb = false;
			}

			free(temp);
			cm->freePitch(A);
			cm->freePitch(B);
			cm->freePitch(C);
		}
		passeds.push_back(gb);
		PRINT_STATUS
	}

	// Test scale and offset.
	{
		test_names.push_back("Pitch scale and offset");
		gb = true;
		float gf;
		int n_reps = 10;
		for ( unsigned j=0; gb && j<2; ++j ) {
			svlCudaPitch *A = cm->allocPitch(ws[j], hs[j]);
			
			for ( int k=0; gb && k<n_reps; ++k ) {
				cm->pitchToGPU(A, datas[j]);
				float scale = 2.0f * float(rand())/float(RAND_MAX) - 1.0f,
					  offset = 10.0f * float(rand())/float(RAND_MAX) - 1.0f;
				// Performs <A = scale*A + offset>.
				svlCudaMath::scaleOffset(A, scale, offset);

				// Compare with gold result.
				cm->pitchFromGPU(tests[j], A);
				float max_d = 0.0, mean_d = 0.0;
				for ( unsigned i=0; i<szs[j]; ++i ) {
					gf = scale * datas[j][i] + offset;
					mean_d += fabs( gf - tests[j][i] );
					max_d = max( max_d, fabs( gf - tests[j][i] ) );
				}
				mean_d /= float(A->size());
				//printf("mean: %0.9f; max: %0.9f\n", mean_d, max_d);
				if ( mean_d >= 1e-7 || max_d >= 1e-6 )
					gb = false;
			}

			cm->freePitch(A);
		}
		passeds.push_back(gb);
		PRINT_STATUS
	}

	// Test decimation.
	{
		test_names.push_back("Pitch decimation");
		gb = true;
		for ( unsigned j=0; gb && j<2; ++j ) {
			svlCudaPitch *A = cm->allocPitch(ws[j], hs[j]);
			svlCudaPitch *B = cm->allocPitch(ws[j], hs[j]);
			cm->pitchToGPU(A, datas[j]);

			for ( int dx=2; gb && dx<=8; dx*=2 ) {
				for ( int dy=2; gb && dy<=8; dy*=2 ) {
					// Performs decimation.
					svlCudaMath::decimate(A, B, dx, dy);
					cm->pitchFromGPU(tests[j], B);

					// Compute gold result.
					float max_d = 0.0, sum_d = 0.0;
					for ( unsigned r=0; r<hs[j]/dy; ++r ) {
						for ( unsigned c=0; c<ws[j]/dx; ++c ) {
							gf = datas[j][ r*dy*ws[j] + c*dx ] - tests[j][ r*ws[j] + c ];
							sum_d += fabs(gf);
							max_d = max( max_d, fabs(gf) );
						}
					}

					if ( sum_d >= TRANFER_TOL || max_d >= TRANFER_TOL )
						gb = false;
				}
			}

			cm->freePitch(A);
			cm->freePitch(B);
		}
		passeds.push_back(gb);
		PRINT_STATUS
	}

	// Test quadrature add.
	{
		test_names.push_back("Pitch quadrature add");
		gb = true;
		for ( unsigned j=0; gb && j<2; ++j ) {
			svlCudaPitch *A = cm->allocPitch(ws[j], hs[j]);
			svlCudaPitch *B = cm->allocPitch(ws[j], hs[j]);
			cm->pitchToGPU(A, datas[j]);
			cm->pitchToGPU(B, datas_2[j]);

			// Performs <A = sqrt(A.^2 + B.^2)>.
			svlCudaMath::addQuadrature(A, B);
			cm->pitchFromGPU(tests[j], A);

			float max_d = 0.0f;
			for ( unsigned i=0; i<szs[j]; ++i ) {
				gf = sqrt( datas[j][i]*datas[j][i] + datas_2[j][i]*datas_2[j][i] );
				max_d = max( max_d, fabs( gf - tests[j][i] ) );
			}
			if ( max_d >= 1e-6 )
				gb = false;

			cm->freePitch(A);
			cm->freePitch(B);
		}
		passeds.push_back(gb);
		PRINT_STATUS
	}

	// Test 32-by-32 mean.
	{
		test_names.push_back("32-by-32 mean");
		gb = true;
		float max_d;
		for ( unsigned j=0; gb && j<2; ++j ) {
			svlCudaPitch *A = cm->allocPitch(ws[j], hs[j]),
						 *B = cm->allocPitch(ws[j], hs[j]),
						 *T = cm->allocPitch(ws[j], hs[j]);
			cm->pitchToGPU(A, datas[j]);

			svlCudaMath::mean32x32(A, B, T);
			cm->pitchFromGPU(tests[j], B);

			mean_gold(datas[j], verfs[j], ws[j], hs[j], 32, 32);
			
			max_d = 0.0f;			
			for ( unsigned i=0; i<szs[j]; ++i )
				max_d = max( max_d, fabs( verfs[j][i] - tests[j][i] ) );
			if ( max_d >= 1e-5 )
				gb = false;

			cm->freePitch(A);
			cm->freePitch(B);
			cm->freePitch(T);
		}
		passeds.push_back(gb);
		PRINT_STATUS
	}
#endif


#ifdef TEST_MINMAX
	// *** Test min/max functions.
	// Test 4-by-4 max.
	{
		test_names.push_back("4-by-4 min/max");
		gb = true;
		for ( unsigned j=0; gb && j<2; ++j ) {
			svlCudaPitch *A = cm->allocPitch(ws[j], hs[j]);
			svlCudaPitch *MINP = cm->allocPitch(ws[j], hs[j]);
			svlCudaPitch *MAXP = cm->allocPitch(ws[j], hs[j]);
			cm->pitchToGPU(A, datas[j]);

			// Minimum
			svlCudaMath::minMax4x4(A, MINP, NULL);
			cm->pitchFromGPU(tests[j], MINP);
			float max_d = 0.0;
			minMax_gold(datas[j], verfs[j], NULL, ws[j], hs[j], 4, 4);
			for ( unsigned i=0; i<szs[j]; ++i )
				max_d = max( max_d, fabs( verfs[j][i] - tests[j][i] ) );

			// Maximum
			svlCudaMath::minMax4x4(A, NULL, MAXP);
			cm->pitchFromGPU(tests[j], MAXP);
			minMax_gold(datas[j], NULL, verfs[j], ws[j], hs[j], 4, 4);
			for ( unsigned i=0; i<szs[j]; ++i )
				max_d = max( max_d, fabs( verfs[j][i] - tests[j][i] ) );

			if ( max_d >= 1e-9 )
				gb = false;

			cm->freePitch(A);
			cm->freePitch(MINP);
			cm->freePitch(MAXP);
		}
		passeds.push_back(gb);
		PRINT_STATUS
	}

	// Test 32-by-32 max.
	{
		test_names.push_back("32-by-32 min/max");
		gb = true;
		for ( unsigned j=0; gb && j<2; ++j ) {
			svlCudaPitch *A = cm->allocPitch(ws[j], hs[j]),
						 *MINP = cm->allocPitch(ws[j], hs[j]),
						 *MAXP = cm->allocPitch(ws[j], hs[j]),
						 *TEMP = cm->allocPitch(ws[j], hs[j]);
			cm->pitchToGPU(A, datas[j]);

			// Minimum
			svlCudaMath::minMax32x32(A, TEMP, MINP, NULL);
			cm->pitchFromGPU(tests[j], MINP);
			float max_d = 0.0;
			minMax_gold(datas[j], verfs[j], NULL, ws[j], hs[j], 32, 32);
			for ( unsigned i=0; i<szs[j]; ++i )
				max_d = max( max_d, fabs( verfs[j][i] - tests[j][i] ) );
			cm->pitchToGPU(MINP, verfs[j]);

			// Maximum
			svlCudaMath::minMax32x32(A, TEMP, NULL, MAXP);
			cm->pitchFromGPU(tests[j], MAXP);
			minMax_gold(datas[j], NULL, verfs[j], ws[j], hs[j], 32, 32);
			for ( unsigned i=0; i<szs[j]; ++i )
				max_d = max( max_d, fabs( verfs[j][i] - tests[j][i] ) );

			if ( max_d >= 1e-9 )
				gb = false;

			cm->freePitch(A);
			cm->freePitch(MINP);
			cm->freePitch(MAXP);
			cm->freePitch(TEMP);
		}
		passeds.push_back(gb);
		PRINT_STATUS
	}
#endif


#ifdef TEST_INTEGRAL
	// *** Test integration functions.
	// Test line integration
	{
		test_names.push_back("Line integrals");
		gb = true;

		int w = 8192;
		svlCudaLine *ln = cm->allocLine(w),
					*integ = cm->allocLine(w);
		cm->lineToGPU(ln, datas[0]);
		float max_d = 0.0f;
		for ( int i=0; i<2; ++i ) {
			for ( int dim=32; dim<=1024; dim*=2 ) {
				bool do_cumsum = i==0;
				svlCudaMath::integralLine(ln, integ, w, dim, do_cumsum);
				cm->lineFromGPU(tests[0], integ);
				
				// Compute gold.
				integralLine_gold(datas[0], verfs[0], w, dim, do_cumsum);
				for ( unsigned j=0; j<w; ++j )
					max_d = max( max_d, fabs( tests[0][j] - verfs[0][j] ) );
				
				if ( max_d >= 1e-3 )
					gb = false;
			}
		}
		cm->freeLine(ln);
		cm->freeLine(integ);
		passeds.push_back(gb);
		PRINT_STATUS
	}

	// Test pitch block integration
	{
		test_names.push_back("Pitch block integrals");
		gb = true;

		float max_d = 0.0f, tol;
		for ( unsigned j=0; gb && j<2; ++j ) {
			svlCudaPitch *A = cm->allocPitch(ws[j], hs[j]),
						 *INTEG = cm->allocPitch(ws[j]+1, hs[j]+1),
						 *INTEG2 = cm->allocPitch(ws[j]+1, hs[j]+1),
						 *TEMP1 = cm->allocPitch(ws[j]+1, hs[j]+1),
						 *TEMP2 = cm->allocPitch(ws[j]+1, hs[j]+1);
			cm->pitchToGPU(A, datas[j]);

			int isz = (ws[j]+1)*(hs[j]+1);
			float *verf = (float*)malloc(sizeof(float)*isz),
				  *integ = (float*)malloc(sizeof(float)*isz),
				  *integ2 = (float*)malloc(sizeof(float)*isz);

			for ( int dim=32; gb && dim<=512; dim*=2 ) {
				tol = 1e-5 * float(dim)*float(dim);
				//printf("tol: %0.9f\n", tol);
				svlCudaMath::integralBlock(A, INTEG, TEMP1, INTEG2, TEMP2, ws[j], hs[j], dim);
				integralBlock_gold(datas[j], integ, integ2, ws[j], hs[j], dim);

				cm->pitchFromGPU(verf, INTEG);
				max_d = 0.0f;
				for ( int i=0; i<isz; ++i )
					max_d = max( max_d, fabs( verf[i] - integ[i] ) );
				//printf("max: %0.9f\n", max_d);
				if ( max_d > tol )
					gb = false;

				cm->pitchFromGPU(verf, INTEG2);
				max_d = 0.0f;
				for ( int i=0; i<isz; ++i )
					max_d = max( max_d, fabs( verf[i] - integ2[i] ) );
				//printf("max: %0.9f\n", max_d);
				if ( max_d > tol )
					gb = false;
			}

			free(integ);
			free(integ2);
			cm->freePitch(A);
			cm->freePitch(INTEG);
			cm->freePitch(INTEG2);
			cm->freePitch(TEMP1);
			cm->freePitch(TEMP2);
		}
		passeds.push_back(gb);
		PRINT_STATUS
	}

	// Test pitch block row integration
	{
		test_names.push_back("Pitch block row integrals");
		gb = true;

		float max_d = 0.0f, tol;
		for ( unsigned j=0; gb && j<2; ++j ) {
			svlCudaPitch *A = cm->allocPitch(ws[j], hs[j]),
						 *INTEG = cm->allocPitch(ws[j]+1, hs[j]),
						 *INTEG2 = cm->allocPitch(ws[j]+1, hs[j]);
			cm->pitchToGPU(A, datas[j]);

			int isz = (ws[j]+1)*hs[j];
			float *verf = (float*)malloc(sizeof(float)*isz),
				  *integ = (float*)malloc(sizeof(float)*isz),
				  *integ2 = (float*)malloc(sizeof(float)*isz);

			for ( int dim=32; gb && dim<=512; dim*=2 ) {
				tol = 1e-6 * float(dim);
				svlCudaMath::rowIntegralBlock(A, INTEG, INTEG2, ws[j], hs[j], dim);
				rowIntegralBlock_gold(datas[j], integ, integ2, ws[j], hs[j], dim);

				cm->pitchFromGPU(verf, INTEG);
				max_d = 0.0f;
				for ( int i=0; i<isz; ++i )
					max_d = max( max_d, fabs( verf[i] - integ[i] ) );
				if ( max_d > tol )
					gb = false;

				cm->pitchFromGPU(verf, INTEG2);
				max_d = 0.0f;
				for ( int i=0; i<isz; ++i )
					max_d = max( max_d, fabs( verf[i] - integ2[i] ) );
				if ( max_d > tol )
					gb = false;
			}

			free(integ);
			free(integ2);
			cm->freePitch(A);
			cm->freePitch(INTEG);
			cm->freePitch(INTEG2);
		}
		passeds.push_back(gb);
		PRINT_STATUS
	}
#endif


#ifdef TEST_CONVOLUTION
	// *** Test convolution functions.
	// Test plain two-dimensional convolution.
	{
		test_names.push_back("'Same' and 'full' convolution");
		gb = true;

		svlCudaConvolution *cc;
		// Generate random patch.
		const int pw = 4 + (rand()%12), ph = 4 + (rand()%12);
		//const int pw = 16, ph = 16;
		int psz = pw*ph;
		float *pat = (float*)malloc(sizeof(float)*psz);
		IplImage *pati = cvCreateImage(cvSize(pw,ph), IPL_DEPTH_32F, 1);
		for ( unsigned i=0; i<psz; ++i ) {
			pat[i] = 2.0f*( float(rand())/float(RAND_MAX) ) + 1.0f;
			CV_IMAGE_ELEM(pati, float, i/pw, i%pw) = pat[i];
		}

		for ( unsigned j=0; gb && j<2; ++j ) {
			svlCudaPitch *A = cm->allocPitch(ws[j], hs[j]),
						 *SAME = cm->allocPitch(ws[j], hs[j]),
						 *FULL = cm->allocPitch(ws[j]+pw-1, hs[j]+ph-1);
			cm->pitchToGPU(A, datas[j]);

			float *full_verf = (float*)malloc(sizeof(float)*( ws[j]+pw-1 )*( hs[j]+ph-1 )),
				  *full_test = (float*)malloc(sizeof(float)*( ws[j]+pw-1 )*( hs[j]+ph-1 ));

			// Compute results with float* arguments.
			cc->cudaMatchTemplate(A, pat, pw, ph, SAME, CUDA_TM_CCORR_SAME);
			conv2d_same(datas[j], ws[j], hs[j], pat, pw, ph, verfs[j]);
			cc->cudaMatchTemplate(A, pat, pw, ph, FULL, CUDA_TM_CCORR_FULL);
			conv2d_full(datas[j], ws[j], hs[j], pat, pw, ph, full_verf);

			cm->pitchFromGPU(tests[j], SAME);
			cm->pitchFromGPU(full_test, FULL);
			float max_d = 0.0f;
			for ( unsigned i=0; i<ws[j]*hs[j]; ++i )
				max_d = max( max_d, fabs( tests[j][i] - verfs[j][i] ) );
			if ( PRINT_TOL ) printf("max: %0.9f\n", max_d);
			if ( max_d >= 2e-4 )
				gb = false;

			max_d = 0.0f;
			for ( unsigned i=0; i<(ws[j]+pw-1)*(hs[j]+ph-1); ++i )
				max_d = max( max_d, fabs( full_test[i] - full_verf[i] ) );
			if ( PRINT_TOL ) printf("max: %0.9f\n", max_d);
			if ( max_d >= 2e-4 )
				gb = false;

			// Compute results with IplImage* arguments.
			cc->cudaMatchTemplate(A, pati, SAME, CUDA_TM_CCORR_SAME);
			cc->cudaMatchTemplate(A, pati, FULL, CUDA_TM_CCORR_FULL);
			cm->pitchFromGPU(tests[j], SAME);
			cm->pitchFromGPU(full_test, FULL);

			max_d = 0.0f;
			for ( unsigned i=0; i<ws[j]*hs[j]; ++i )
				max_d = max( max_d, fabs( tests[j][i] - verfs[j][i] ) );
			if ( PRINT_TOL ) printf("max: %0.9f\n", max_d);
			if ( max_d >= 2e-4 )
				gb = false;

			max_d = 0.0f;
			for ( unsigned i=0; i<(ws[j]+pw-1)*(hs[j]+ph-1); ++i )
				max_d = max( max_d, fabs( full_test[i] - full_verf[i] ) );
			if ( PRINT_TOL ) printf("max: %0.9f\n", max_d);
			if ( max_d >= 2e-4 )
				gb = false;

			free(full_verf);
			free(full_test);
			cm->freePitch(A);
			cm->freePitch(SAME);
			cm->freePitch(FULL);
		}
		passeds.push_back(gb);
		PRINT_STATUS
	}

	// Test normalized correlation coefficient calculation.
	{
		test_names.push_back("Normalized correlation coefficient");
		gb = true;

		svlCudaConvolution *cc;
		// Generate random patch.
		//const int pw = 4 + (rand()%12), ph = 4 + (rand()%12);
		const int pw = 16, ph = 16;
		int psz = pw*ph;
		float *pat = (float*)malloc(sizeof(float)*psz);
		IplImage *pati = cvCreateImage(cvSize(pw,ph), IPL_DEPTH_32F, 1);
		for ( unsigned i=0; i<psz; ++i ) {
			pat[i] = 2.0f*( float(rand())/float(RAND_MAX) ) + 1.0f;
			CV_IMAGE_ELEM(pati, float, i/pw, i%pw) = pat[i];
		}

		for ( unsigned j=0; gb && j<2; ++j ) {
			svlCudaPitch *A = cm->allocPitch(ws[j], hs[j]),
						 *CCOEFF = cm->allocPitch(ws[j], hs[j]);
			cm->pitchToGPU(A, datas[j]);

			// Compute results with float* argument.
			cc->cudaMatchTemplate(A, pat, pw, ph, CCOEFF, CUDA_TM_CCOEFF_NORMED_SAME);
			//svlCudaConvolution::ccoeffNormed(A, pat, pw, ph, CCOEFF);
			ccoeff_normed_gold(datas[j], ws[j], hs[j], pat, pw, ph, verfs[j]);
			
			cm->pitchFromGPU(tests[j], CCOEFF);
			float max_d = 0.0f;
			for ( unsigned i=0; i<ws[j]*hs[j]; ++i )
				max_d = max( max_d, fabs( tests[j][i] - verfs[j][i] ) );
			if ( PRINT_TOL ) printf("max: %0.9f\n", max_d);
			if ( max_d >= 5e-6 )
				gb = false;

			// Compute results with IplImage* argument.
			cc->cudaMatchTemplate(A, pati, CCOEFF, CUDA_TM_CCOEFF_NORMED_SAME);
			//svlCudaConvolution::ccoeffNormed(A, pati, CCOEFF);
			
			cm->pitchFromGPU(tests[j], CCOEFF);
			max_d = 0.0f;
			for ( unsigned i=0; i<ws[j]*hs[j]; ++i )
				max_d = max( max_d, fabs( tests[j][i] - verfs[j][i] ) );
			if ( PRINT_TOL ) printf("max: %0.9f\n", max_d);
			if ( max_d >= 5e-6 )
				gb = false;

			cm->freePitch(A);
			cm->freePitch(CCOEFF);
		}

		passeds.push_back(gb);
		PRINT_STATUS
	}
#endif


#ifdef TEST_FILTERING
	// Test Sobel filter.
	{
		test_names.push_back("Sobel filter");
		gb = true;

		for ( unsigned j=0; gb && j<2; ++j ) {
			svlCudaPitch *A = cm->allocPitch(ws[j], hs[j]),
						 *E = cm->allocPitch(ws[j], hs[j]),
						 *T1 = cm->allocPitch(ws[j], hs[j]),
						 *T2 = cm->allocPitch(ws[j], hs[j]);
			cm->pitchToGPU(A, datas[j]);

			// Computer results and gold result.
			svlCudaFilter::sobel(A, E, T1, T2);
			sobel_gold(datas[j], ws[j], hs[j], verfs[j]);
			
			cm->pitchFromGPU(tests[j], E);
			float max_d = 0.0f;
			for ( unsigned i=0; i<ws[j]*hs[j]; ++i )
				max_d = max( max_d, fabs( tests[j][i] - verfs[j][i] ) );
			//printf("max: %0.9f\n", max_d);
			if ( max_d >= 1e-6 )
				gb = false;

			cm->freePitch(A);
			cm->freePitch(E);
			cm->freePitch(T1);
			cm->freePitch(T2);
		}

		passeds.push_back(gb);
		PRINT_STATUS
	}
#endif


#ifdef TEST_DECISION_TREES
	// Test decision tree.
	{
		test_names.push_back("Decision trees");
		gb = true;

		const int feature_len = 1200, n = 100;
		svlCudaLine *res = cm->allocLine(n),
					*invalid = cm->allocLine(n);
		// Create random features.
		float *f = (float*)malloc(sizeof(float)*feature_len*n);
		for ( int i=0; i<feature_len*n; ++i )
			f[i] = 2.0f*( float(rand())/float(RAND_MAX) ) - 1.0f;
		float *ref = (float*)malloc(sizeof(float)*n),
			  *test = (float*)malloc(sizeof(float)*n);

		// Declare decision tree and special feature line.
		svlCudaDTree15 dtree("../tests/input/cuda_tree.dtree");
		svlCudaFeatureLine feats(feature_len, dtree.getNtrees());
		feats.alloc(n); // Allocate space.
		cm->lineToGPU(feats.features, f);

		// Run device computation and reference computation.
		feats.evaluate(&dtree, invalid, res);
		dtree.hostEvaluate(f, NULL, ref, feature_len, feature_len, n);
		cm->lineFromGPU(test, res);

		float max_d = 0.0f;
		for ( unsigned i=0; i<n; ++i )
			max_d = max( max_d, fabs( ref[i] - test[i] ) );
		//printf("max: %0.9f\n", max_d);
		if ( max_d > 5e-5 )
			gb = false;

		free(f);
		free(ref);
		free(test);
		cm->freeLine(res);
		cm->freeLine(invalid);

		passeds.push_back(gb);
		PRINT_STATUS
	}
#endif

#ifdef TEST_SLIDING_WINDOW
	// Test sliding window feature extractor.
	{
		test_names.push_back("Sliding window feature extraction");
		gb = true;

		const char *dict = "../tests/input/testDict45.xml";
		int del_x = 4, del_y = 4;
		double del_scale = 1.2;
		double gd;
		for ( unsigned j=0; gb && j<2; ++j ) {
			vector<IplImage*> images;
			IplImage *img = cvCreateImage(cvSize(ws[j], hs[j]), IPL_DEPTH_32F, 1),
					 *edge = cvCreateImage(cvSize(ws[j], hs[j]), IPL_DEPTH_32F, 1),
					 *dep = cvCreateImage(cvSize(ws[j], hs[j]), IPL_DEPTH_32F, 1);
			for ( int r=0; r<hs[j]; ++r ) {
				memcpy( &img->imageData[ img->widthStep*r ], &datas[j][ r*ws[j] ], sizeof(float)*ws[j] );
				memcpy( &edge->imageData[ img->widthStep*r ], &datas_2[j][ r*ws[j] ], sizeof(float)*ws[j] );
				for ( int c=0; c<ws[j]; ++c )
					CV_IMAGE_ELEM(dep, float, r, c) = CV_IMAGE_ELEM(img, float, r, c) + CV_IMAGE_ELEM(edge, float, r, c);
			}
			images.push_back(img);
			images.push_back(edge);
			images.push_back(dep);

			// Declare feature extractors.
			svlCudaSlidingWindowDetector cuda_det(ws[j], hs[j]);
			cuda_det.readDictionary(dict, false);
			cuda_det.setDeltas(del_x, del_y, del_scale);
			cuda_det.padAndCachePatches();
			svlSlidingWindowDetector det("thing", ws[j], hs[j]);
			//det.setFeatureExtractor( svlFeatureExtractor::getInstance(dict) );
			det.load(dict);

			// Get features.
			ScaleLocations locs, clocs;
			ScaleFeaturesD feats;
			ScaleFeatures cfeats;
			cuda_det.classifyImage(images, NULL, &clocs, &cfeats);
			det.computeFeatures(images, feats, locs);
			/*printf("size: %d,%d > %d,%d\n", clocs.size(), cfeats.size(), clocs.size() > 0 ? clocs[0].size() : -1, cfeats.size() > 0 ? cfeats[0].size() : -1 );
			printf("size: %d,%d > %d,%d\n", locs.size(), feats.size(), locs.size() > 0 ? locs[0].size() : -1, feats.size() > 0 ? feats[0].size() : -1 );
			scaleLocsFeats2File("F_cuda.out", clocs, cfeats);
			scaleLocsFeatsD2File("F_gold.out", locs, feats);*/

			// Compute location intersection to compare features.
			double max_d = 0.0f, mean_d = 0.0f;
			unsigned total_sz = 0;
			vector<int> lpts, cpts, lix, cix;
			for ( unsigned i=0; i<locs.size() && i<clocs.size(); ++i ) {
				sub2ind(cvSize(ws[j],hs[j]), locs[i], lpts);
				sortix(lpts, lix);
				sub2ind(cvSize(ws[j],hs[j]), clocs[i], cpts);
				sortix(cpts, cix);

				// Compute intersection.
				vector<int>::const_iterator lit = lix.begin(), cit = cix.begin();
				unsigned ii = 0;
				while ( lit != lix.end() && cit != cix.end() ) {
					if ( lpts[*lit] < cpts[*cit] ) {
						++lit;
					} else if ( cpts[*cit] < lpts[*lit] ) {
						++cit;
					} else {
						// Accumulate difference.
						++total_sz;
						for ( unsigned k=0; k<feats[i][*lit].size(); ++k ) {
							gd = fabs( feats[i][*lit][k] - double(cfeats[i][*cit][k]) );
							mean_d += gd;
							max_d = max( max_d, gd );
						}
						++lit;
						++cit;
					}
				}
			}
			mean_d /= float(total_sz*feats[0][0].size());
			//printf("total: %d; max: %0.9f; mean: %0.9f\n", total_sz, max_d, mean_d);
			if ( total_sz < 1000 || max_d >= 1e-4 || mean_d >= 1e-6 )
				gb = false;
		}
		passeds.push_back(gb);
		PRINT_STATUS
	}
#endif


	int n_passed = 0;
	for ( unsigned i=0; i<test_names.size(); ++i )
		if ( passeds[i] )
			++n_passed;
	printf("\nPassed %d; Failed %d\n", n_passed, int(test_names.size())-n_passed);
	// Print failed tests again.
	if ( n_passed != int(test_names.size()) ) {
		printf("Failed:\n");
		for ( unsigned i=0; i<test_names.size(); ++i ) {
			if ( !passeds[i] )
				printf("  %s\n", test_names[i]);
		}
	}

	printf("\n");
	cm->printAllAllocations();
#endif
	return 0;
}

// Compute min/max of variable size.
void minMax_gold(const float *data, float *minf, float *maxf, int w, int h, int mw, int mh)
{
	float gf;
	for ( unsigned r=0; r<h; ++r ) {
		for ( unsigned c=0; c<w; ++c ) {
			gf = data[ r*w + c ];
			if ( minf ) {
				for ( int rr=0; rr<mh && r+rr < h; ++rr )
					for ( int cc=0; cc<mw && c+cc < w; ++cc )
						gf = min( gf, data[ ( r + rr )*w + c + cc ] );
				minf[ r*w + c ] = gf;
			}
			if ( maxf ) {
				for ( int rr=0; rr<mh && r+rr < h; ++rr )
					for ( int cc=0; cc<mw && c+cc < w; ++cc )
						gf = max( gf, data[ ( r + rr )*w + c + cc ] );
				maxf[ r*w + c ] = gf;
			}
		}
	}
}

// Compute mean of variable size.
void mean_gold(const float *data, float *outf, int w, int h, int mw, int mh)
{
	float gf, fac = 1.0f/float(mw*mh);
	for ( int r=0; r<h; ++r ) {
		for ( int c=0; c<w; ++c ) {
			gf = 0.0f;
			for ( int i=0; i<mh && r + i<h; ++i )
				for ( int j=0; j<mw && c + j<w; ++j )
					gf += data[ (r+i)*w + c + j ];
			outf[ r*w + c ] = fac * gf;
		}
	}
}

// Compute integral line.
void integralLine_gold(const float *data, float *integ, int w, int dim, bool do_cumsum)
{
	float gf;
	for ( int i=0; i<w; i+=dim ) {
		gf = 0.0f;
		for ( int j=i; j<i+dim; ++j ) {
			integ[j] = gf + ( do_cumsum ? data[j] : 0.0f );
			gf += data[j];
		}
	}
}

// Compute integral block.
void integralBlock_gold(const float *data, float *integ, float *integ2, int w, int h, int dim)
{
	float gf, gf2;
	memset(integ, 0, (w+1)*(h+1)*sizeof(float));
	memset(integ2, 0, (w+1)*(h+1)*sizeof(float));
	for ( int rr=0; rr<(int)ceil((float)h/dim); ++rr ) {
		for ( int cc=0; cc<(int)ceil((float)w/dim); ++cc ) {
			for ( int r=rr*dim; r<(rr+1)*dim && r<h; ++r ) {
				gf = gf2 = 0.0f;
				for ( int c=cc*dim; c<(cc+1)*dim && c<w; ++c ) {
					gf += data[ r*w + c ];
					gf2 += data[ r*w + c ] * data[ r*w + c ];
					integ[ (r+1)*(w+1) + c + 1 ] = gf + ( r > rr*dim ? integ[ r*(w+1) + c + 1 ] : 0.0f );
					integ2[ (r+1)*(w+1) + c + 1 ] = gf2 + ( r > rr*dim ? integ2[ r*(w+1) + c + 1 ] : 0.0f );
				}
			}
		}
	}
}

// Compute row integral block.
void rowIntegralBlock_gold(const float *data, float *integ, float *integ2, int w, int h, int dim)
{
	float gf, gf2;
	memset(integ, 0, (w+1)*h*sizeof(float));
	memset(integ2, 0, (w+1)*h*sizeof(float));
	for ( int r=0; r<h; ++r ) {
		for ( int cc=0; cc<(int)ceil((float)w/dim); ++cc ) {
			gf = gf2 = 0.0f;
			for ( int c=cc*dim; c<(cc+1)*dim && c<w; ++c ) {
				integ[ r*(w+1) + c + 1 ] = gf += data[ r*w + c ];
				integ2[ r*(w+1) + c + 1 ] = gf2 += data[ r*w + c ] * data[ r*w + c ];
			}
		}
	}
}


void scaleLocsFeats2File(const char *fname, const ScaleLocations &locs, const ScaleFeatures &feats)
{
	//printf("Saving info to '%s' ...", fname);
	FILE *fo = fopen(fname,"wb");
	int gi, feature_stride = feats[0][0].size();
	fwrite(&feature_stride,sizeof(int),1,fo);
	for ( int i=0; i<int(locs.size()); ++i ) {
		fwrite(&(gi=locs[i].size()),sizeof(int),1,fo);
		// Write locations.
		for ( vector<CvPoint>::const_iterator c=locs[i].begin(); c!=locs[i].end(); ++c ) {
			fwrite(&c->x,sizeof(int),1,fo);
			fwrite(&c->y,sizeof(int),1,fo);
		}
		// Write features.
		for ( vector<vector<float> >::const_iterator fv=feats[i].begin(); fv!=feats[i].end(); ++fv ) {
			for ( vector<float>::const_iterator f=fv->begin(); f!=fv->end(); ++f ) {
				fwrite(&(*f),sizeof(float),1,fo);
			}
		}
	}
	fclose(fo);
	//printf(" done.\n");
}

void scaleLocsFeatsD2File(const char *fname, const ScaleLocations &locs, const ScaleFeaturesD &feats)
{
	//printf("Saving info to '%s' ...", fname);
	FILE *fo = fopen(fname,"wb");
	int gi, feature_stride = feats[0][0].size();
	fwrite(&feature_stride,sizeof(int),1,fo);
	for ( int i=0; i<int(locs.size()); ++i ) {
		fwrite(&(gi=locs[i].size()),sizeof(int),1,fo);
		// Write locations.
		for ( vector<CvPoint>::const_iterator c=locs[i].begin(); c!=locs[i].end(); ++c ) {
			fwrite(&c->x,sizeof(int),1,fo);
			fwrite(&c->y,sizeof(int),1,fo);
		}
		// Write features.
		for ( vector<vector<double> >::const_iterator fv=feats[i].begin(); fv!=feats[i].end(); ++fv ) {
			for ( vector<double>::const_iterator f=fv->begin(); f!=fv->end(); ++f ) {
				fwrite(&(*f),sizeof(double),1,fo);
			}
		}
	}
	fclose(fo);
	//printf(" done.\n");
}

// Subscript to indices conversion.
void sub2ind(CvSize sz, const vector<CvPoint> &pts, vector<int> &ret)
{
	ret.resize(pts.size());
	for ( unsigned i=0; i<pts.size(); ++i )
		ret[i] = pts[i].y*sz.width + pts[i].x;
}

// Return sort indices.
void sortix(const vector<int> &vec, vector<int> &ix)
{
	vector<pair<int,int> > tosort;
	tosort.reserve(vec.size());
	for ( unsigned i=0; i<vec.size(); ++i )
		tosort.push_back(pair<int,int>(vec[i],i));
	
	sort(tosort.begin(), tosort.end());
	ix.resize(vec.size());
	for ( unsigned i=0; i<tosort.size(); ++i )
		ix[i] = tosort[i].second;
}

// Returns a convolution result.
void conv2d_same(const float *data, int dw, int dh, const float *pat, int pw, int ph, float *res)
{
	float gf;
	for ( int r=0; r<dh; ++r ) {
		for ( int c=0; c<dw; ++c ) {
			gf = 0.0f;
			for ( int i=0; i<ph && r+i<dh; ++i )
				for ( int j=0; j<pw && c+j<dw; ++j )
					gf += data[ (r+i)*dw + c+j ] * pat[ i*pw + j ];
			res[ r*dw + c ] = gf;
		}
	}
}

// The above, with "full."
void conv2d_full(const float *data, int dw, int dh, const float *pat, int pw, int ph, float *res)
{
	float gf;
	for ( int r=0; r<dh+ph-1; ++r ) {
		for ( int c=0; c<dw+pw-1; ++c ) {
			gf = 0.0f;
			for ( int i=max(0, r-ph+1); i<min(dh, r+1); ++i )
				for ( int j=max(0, c-pw+1); j<min(dw, c+1); ++j )
					gf += data[ i*dw + j ] * pat[ ( i - (r-ph+1) )*pw + j - (c-pw+1) ];
			res[ r*(dw+pw-1) + c ] = gf;
		}
	}
}

// Normalized correlation coefficient.
void ccoeff_normed_gold(const float *data, int dw, int dh, const float *pat, int pw, int ph, float *res)
{
	// Compute mean and std of patch.
	int psz = pw*ph;
	float mean = 0.0f, std = 0.0f;
	for ( int i=0; i<psz; ++i ) {
		mean += pat[i];
		std += pat[i] * pat[i];
	}
	mean /= float(psz);
	std = sqrtf( std - float(psz)*mean*mean );
	
	float gf, sum, sum2, d;
	for ( int r=0; r<dh; ++r ) {
		for ( int c=0; c<dw; ++c ) {
			gf = sum = sum2 = 0.0f;
			for ( int i=0; i<ph && r+i<dh; ++i ) {
				for ( int j=0; j<pw && c+j<dw; ++j ) {
					d = data[ (r+i)*dw + c+j ];
					gf += d * pat[ i*pw + j ];
					sum += d;
					sum2 += d * d;
				}
			}
			res[ r*dw + c ] = ( gf - mean * sum ) / sqrtf( sum2 - sum*sum/float(psz) ) / std;
		}
	}
}

// Sobel filter.
void sobel_gold(const float *data, int dw, int dh, float *edge)
{
	int N = dw*dh;
	float *t1 = (float*)malloc(sizeof(float)*N),
		  *t2 = (float*)malloc(sizeof(float)*N);

	// Do col [1 2 1] on data to t1, and col [1 0 -1]' on data to t2.
	for ( int c=0; c<dw; ++c ) {
		for ( int r=0; r<dh; ++r ) {
			int ix = r*dw + c;
			t1[ix] = ( r > 0 ? data[ix - dw] : 0.0f ) + 2.0f * data[ix] + ( r < dh-1 ? data[ix + dw] : 0.0f );
			t2[ix] = ( r > 0 ? data[ix - dw] : 0.0f ) - ( r < dh-1 ? data[ix + dw] : 0.0f );
		}
	}

	// Do row [1 0 -1] on t1 to edge
	for ( int r=0; r<dh; ++r ) {
		for ( int c=0; c<dw; ++c ) {
			int ix = r*dw + c;
			edge[ix] = ( c > 0 ? t1[ix - 1] : 0.0f ) - ( c < dw-1 ? t1[ix + 1] : 0.0f );
		}
	}

	// Do row [1 2 1] on t2 to t1.
	for ( int r=0; r<dh; ++r ) {
		for ( int c=0; c<dw; ++c ) {
			int ix = r*dw + c;
			t1[ix] = ( c > 0 ? t2[ix - 1] : 0.0f ) + 2.0f * t2[ix] + ( c < dw-1 ? t2[ix + 1] : 0.0f );
		}
	}

	// Add t1 to edge in quadrature.
	for ( int r=0; r<dh; ++r ) {
		for ( int c=0; c<dw; ++c ) {
			int ix = r*dw + c;
			edge[ix] = sqrt( edge[ix]*edge[ix] + t1[ix]*t1[ix] );
		}
	}

	free(t1);
	free(t2);
}


