/*****************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    svlCudaDTree.h
** AUTHOR(S):   Paul Baumstarck <pbaumstarck@stanford.edu>
** DESCRIPTION:
**  Implements a Cuda decision tree of depth 3.
**
*****************************************************************************/

#pragma once

#ifndef __SVL_CUDA_DTREE_H
#define __SVL_CUDA_DTREE_H

#include <iostream>
#include <vector>

#include "cv.h"
#include "cxcore.h"

#include "svlCudaCommon.h"
#include "svlCudaMath.h"
#include "svlCudaMemHandler.h"
using namespace std;


// Cuda decision tree classes.

class svlCudaDTree {
protected:
	int n_trees;
	float *tree;
	svlCudaLine *dtree;

public:
	svlCudaDTree();
	svlCudaDTree(const char *file);
	virtual ~svlCudaDTree();

	inline int getNtrees() { return n_trees; }
	virtual void evaluate(svlCudaLine *features, svlCudaLine *invalid, int feature_len, int feature_stride, int n_fvs, int execs_per_block = 100) const {
		svlCudaDTreeEvaluate(dtree, n_trees, features, feature_len, feature_stride, n_fvs, execs_per_block);
	}
	inline void collectResults(svlCudaLine *features, svlCudaLine *res, int n_fvs, int feature_stride) const {
		svlCudaDTreeCollectResults(features, res, n_trees, n_fvs, feature_stride);
	}

	virtual void hostEvaluate(const float *features, float *invalid, float *res, int feature_len, int feature_stride, int n_fvs) const {};
};

class svlCudaDTree15 : public svlCudaDTree {
protected:
	int *treei;
	float *treef;
	svlCudaLine *dtreei, *dtreef;
public:
	svlCudaDTree15();
	svlCudaDTree15(const char *file);
	virtual ~svlCudaDTree15();

	//inline int getNtrees() { return n_trees; }
	virtual void evaluate(svlCudaLine *features, svlCudaLine *invalid, int feature_len, int feature_stride, int n_fvs, int execs_per_block = 100) const {
		svlCudaDTree15Evaluate(dtreei, dtreef, invalid, n_trees, features, feature_len, feature_stride, n_fvs, execs_per_block);
	}
	virtual void hostEvaluate(const float *features, float *invalid, float *res, int feature_len, int feature_stride, int n_fvs) const;
};


// Special line for use with features.

class svlCudaFeatureLine {
public:
	svlCudaLine *features;		// Storage for features.
	int n_features,				// Number of features (not necessarily storage stride).
		feature_stride,			// Storage stride of features on device.
		tree_aligned_stride,	// If stride needs to be bumped up for tree size.
		n;						// Number of allocated vectors.
	bool native_pow_2;			// True if rounded-up stride is equal to original stride, which dictates using cumsum-style summation instead of scan-style.

public:
	svlCudaFeatureLine(int n_features_in, int n_trees) {
		n_features = n_features_in;
		tree_aligned_stride = svlCudaMath::iAlignUpPow2(n_trees);
		feature_stride = max( tree_aligned_stride, n_features );
		native_pow_2 = n_trees == feature_stride;
		n = -1;
		features = NULL;
	}
	virtual ~svlCudaFeatureLine() {
		thisFree();
	}

	inline void alloc(int new_n) {
		thisFree();
		n = new_n;
		features = svlCudaMemHandler::allocLine(feature_stride * n);
	}
	inline void thisFree() {
		if ( features ) {
			svlCudaMemHandler::freeLine(features);
		}
		n = -1;
	}

	inline void evaluate(const svlCudaDTree *tree, svlCudaLine *invalid) {
		svlCudaMemHandler::lineSet(invalid, 0); // Clear invalid bits.
		tree->evaluate(features, invalid, n_features, feature_stride, n);
	}

	inline void evaluate(const svlCudaDTree *tree, svlCudaLine *invalid, svlCudaLine *results) {
		evaluate(tree, invalid);
		// Integrate features to get response of all trees.
		svlCudaMath::integralLine(features, features, n * feature_stride, tree_aligned_stride, native_pow_2, feature_stride);
		// Compact results and read back out.
		tree->collectResults(features, results, n, feature_stride);
	}
};

#endif

